@@ -11,21 +11,20 @@ class Scheduler:
 		self.group = group
 		
 		self.handle = itertools.count()
-		self.event = anyio.create_event()
+		self.event = anyio.Event()
 		self.events = {}
 		
-	async def start(self):
-		await self.group.spawn(self.process)
+	def start(self):
+		self.group.start_soon(self.process)
 	
 	async def process(self):
 		while True:
-			timeout = await self.process_timers()
-			async with anyio.move_on_after(timeout):
+			timeout = self.process_timers()
+			with anyio.move_on_after(timeout):
 				await self.event.wait()
-				self.event = anyio.create_event()
+				self.event = anyio.Event()
 	
-	async def process_timers(self):
-		minimum = None
+	def process_timers(self):
 		current = time.monotonic()
 		items = self.events.copy().items()
 		for handle, (deadline, repeat, function, args) in items:
@@ -33,26 +32,25 @@ class Scheduler:
 				del self.events[handle]
 				if repeat is not None:
 					self.events[handle] = (deadline + repeat, repeat, function, args)
-				await self.group.spawn(function, *args)
-			else:
-				if minimum is None or minimum > deadline - current:
-					minimum = deadline - current
-		return minimum
+				self.group.start_soon(function, *args)
+		
+		timeouts = [event[0] - current for event in self.events.values()]
+		return min(timeouts, default=None)
 	
-	async def schedule(self, function, delay, *args):
+	def schedule(self, function, delay, *args):
 		deadline = time.monotonic() + delay
 		
 		handle = next(self.handle)
 		self.events[handle] = (deadline, None, function, args)
-		await self.event.set()
+		self.event.set()
 		return handle
 	
-	async def repeat(self, function, delay, *args):
+	def repeat(self, function, delay, *args):
 		deadline = time.monotonic() + delay
 		
 		handle = next(self.handle)
 		self.events[handle] = (deadline, delay, function, args)
-		await self.event.set()
+		self.event.set()
 		return handle
 		
 	def remove(self, handle):
@@ -67,5 +65,5 @@ class Scheduler:
 async def create():
 	async with util.create_task_group() as group:
 		scheduler = Scheduler(group)
-		await scheduler.start()
+		scheduler.start()
 		yield scheduler
