@@ -11,7 +11,7 @@ logger = logging.getLogger(__name__)
 class TCPClient:
 	def __init__(self, stream):
 		self.stream = stream
-		self.lock = anyio.create_lock()
+		self.lock = anyio.Lock()
 	
 	async def send(self, data):
 		async with self.lock:
@@ -46,5 +46,5 @@ async def serve(handler, host="", port=0):
 	listener = await anyio.create_tcp_listener(local_host=host, local_port=port)
 	async with listener:
 		async with util.create_task_group() as group:
-			await group.spawn(listener.serve, handle)
+			group.start_soon(listener.serve, handle)
 			yield
